# -*- coding: utf-8 -*-
"""DdecCp2kChargesWorkChain workchain of the AiiDA DDEC plugin"""

from __future__ import absolute_import

from aiida.common import AttributeDict
from aiida.engine import WorkChain, ToContext
from aiida.orm import Dict
from aiida.plugins import CalculationFactory, DataFactory, WorkflowFactory

from aiida_ddec.utils import merge_Dict, extract_core_electrons

Cp2kBaseWorkChain = WorkflowFactory('cp2k.base')  # pylint: disable=invalid-name
DdecCalculation = CalculationFactory('ddec')  # pylint: disable=invalid-name
CifData = DataFactory('cif')  # pylint: disable=invalid-name


class Cp2kDdecWorkChain(WorkChain):
    """A workchain that computes DDEC charges after a single-point
    DFT calculation using Cp2kBaseWorkChain
    """

    @classmethod
    def define(cls, spec):
        """Define workflow specification."""
        super(Cp2kDdecWorkChain, cls).define(spec)

        spec.expose_inputs(Cp2kBaseWorkChain, namespace='cp2k_base')
        spec.expose_inputs(DdecCalculation, namespace='ddec')

        spec.outline(cls.run_cp2k, cls.run_ddec, cls.return_results)

        spec.expose_outputs(Cp2kBaseWorkChain, include=['remote_folder'])
        spec.expose_outputs(DdecCalculation, include=['structure_ddec'])

        spec.exit_code(903, 'ERROR_PARSING_CP2K_OUTPUT', 'Error while parsing CP2K output')
        spec.exit_code(904, 'ERROR_PARSING_DDEC_OUTPUT', 'Error while parsing DDEC output')

    def run_cp2k(self):
        """Modify CP2K input to compute the charge density,
        and run the ENERGY calculation
        """

        param_modify = Dict(
            dict={
                'GLOBAL': {
                    'RUN_TYPE': 'ENERGY'
                },
                'FORCE_EVAL': {
                    'DFT': {
                        'PRINT': {
                            'E_DENSITY_CUBE': {
                                '_': 'ON',
                                'STRIDE': '1 1 1'
                            }
                        }
                    }
                }
            }
        ).store()

        cp2k_base_inputs = AttributeDict(self.exposed_inputs(Cp2kBaseWorkChain, 'cp2k_base'))
        cp2k_base_inputs['cp2k']['parameters'] = merge_Dict(cp2k_base_inputs['cp2k']['parameters'], param_modify)
        cp2k_base_inputs['metadata']['label'] = 'cp2k_energy'
        cp2k_base_inputs['metadata']['call_link_label'] = 'call_cp2k_energy'
        running = self.submit(Cp2kBaseWorkChain, **cp2k_base_inputs)
        self.report('Running Cp2kBaseWorkChain to compute the charge-density')
        return ToContext(cp2k_calc=running)

    def run_ddec(self):
        """Parse the CP2K ouputs (E_DENSITY_CUBE and core electrons),
        generate the DDEC input and submit DDEC calculation.
        """
        # extract number of core electrons from the cp2k output and prepare input
        try:
            core_e = extract_core_electrons(self.ctx.cp2k_calc.outputs.remote_folder)
            ddec_inputs = AttributeDict(self.exposed_inputs(DdecCalculation, 'ddec'))
            ddec_inputs['charge_density_folder'
                       ] = self.ctx.cp2k_calc.outputs.remote_folder.creator.outputs.remote_folder
        except Exception as exc:  # pylint: disable=broad-except
            self.report(f'Encountered exception "{str(exc)}" while parsing CP2K output')
            return self.exit_codes.ERROR_PARSING_CP2K_OUTPUT  # pylint: disable=no-member

        ddec_inputs['parameters'] = merge_Dict(ddec_inputs['parameters'], core_e)
        ddec_inputs['metadata']['call_link_label'] = 'call_ddec_calc'

        # Create the calculation process and launch it
        running = self.submit(DdecCalculation, **ddec_inputs)
        self.report('Running DdecCalculation<{}> to compute point charges from the charge-density'.format(running.pk))
        return ToContext(ddec_calc=running)

    def return_results(self):
        """Return exposed outputs and print the pk of the CifData w/DDEC"""
        try:
            self.out_many(self.exposed_outputs(self.ctx.cp2k_calc, Cp2kBaseWorkChain))
            self.out_many(self.exposed_outputs(self.ctx.ddec_calc, DdecCalculation))
            self.report('DDEC charges computed: CifData<{}>'.format(self.outputs['structure_ddec'].pk))
        except KeyError:
            return self.exit_codes.ERROR_PARSING_DDEC_OUTPUT  # pylint: disable=no-member

        return 0



# -*- coding: utf-8 -*-
"""AiiDA DDEC plugin parser"""
from __future__ import absolute_import
import os
from aiida.parsers.parser import Parser
from aiida.common import NotExistent, OutputParsingError
from aiida.engine import ExitCode
from aiida.plugins import CalculationFactory
from aiida_ddec.utils import xyz2cif

DdecCalculation = CalculationFactory('ddec')  # pylint: disable=invalid-name


class DdecParser(Parser):
    """
    Parser class for parsing output of multiplication.
    """

    # pylint: disable=protected-access
    def parse(self, **kwargs):
        """Parse output structur, store it in database in CifData format."""

        # Check that the retrieved folder is there
        try:
            out_folder = self.retrieved
        except NotExistent:
            self.logger.error('No retrieved folder found')
            return self.exit_codes.ERROR_NO_RETRIEVED_FOLDER

        # Check what is inside the folder
        list_of_files = out_folder._repository.list_object_names()  # pylint: disable=protected-access

        output_file = self.node.process_class._DEFAULT_OUTPUT_FILE

        # We need at least the output file name as defined in calcs.py
        if output_file not in list_of_files:
            raise self.exit_codes.ERROR_NO_OUTPUT_FILE

        finished = False
        with out_folder.open(output_file) as file:
            for line in file.readlines():
                if 'Finished chargemol in' in line:
                    finished = True

        if not finished:
            raise OutputParsingError('Calculation did not finish correctly')

        # Create CifData object from the following the file path returned by xyz2cif
        with out_folder.open('DDEC6_even_tempered_net_atomic_charges.xyz') as handle:
            output_cif = xyz2cif(handle.readlines())
        self.out('structure_ddec', output_cif)

        return ExitCode(0)



# -*- coding: utf-8 -*-
"""Convert XYZ to CIF keeping the charges"""
from __future__ import absolute_import
import re
import tempfile
from math import degrees, acos
from numpy import dot
from numpy.linalg import inv, norm
from six.moves import map

from aiida.plugins import DataFactory

CifData = DataFactory('cif')  # pylint: disable=invalid-name

CIF_HEADER = """data_crystal

_audit_creation_method 'AiiDA-DDEC plugin'

_cell_length_a        {a:8.5f}
_cell_length_b        {b:8.5f}
_cell_length_c        {c:8.5f}

_cell_angle_alpha     {alpha:8.5f}
_cell_angle_beta     {beta:8.5f}
_cell_angle_gamma     {gamma:8.5f}

_symmetry_space_group_name_Hall 'P 1'
_symmetry_space_group_name_H-M  'P 1'

loop_
_symmetry_equiv_pos_as_xyz
 'x,y,z'

loop_
_atom_site_label
_atom_site_type_symbol
_atom_site_fract_x
_atom_site_fract_y
_atom_site_fract_z
_atom_site_charge
"""


def xyzparser(lines):
    """Extract from XYZ files cell, atoms, coordinates, charges.

    :param lines: List of lines of XYZ file, e.g. result of open(filename).readlines()
    """

    # Number of atoms
    natoms = int(lines[0])

    # Cell
    cell_string = lines[1]
    cell_unformatted = re.search(r'\[.*?\]', cell_string).group(0)[1:-1].split(',')
    cell = [list(map(float, re.search(r'\{.*?\}', e).group(0)[1:-1].split())) for e in cell_unformatted]

    # Atoms
    atoms = [i.split()[0] for i in lines[2:natoms + 2]]

    # XYZ coordinates
    xyz = [list(map(float, i.split()[1:4])) for i in lines[2:natoms + 2]]

    # Charges
    charges = [float(i.split()[4]) for i in lines[2:natoms + 2]]

    return cell, atoms, xyz, charges


def cell2abc(cell):
    """Compute norms of cell vectors."""
    return norm(cell[0]), norm(cell[1]), norm(cell[2])


def cell2angles(cell):
    """Return angles alpha, beta, gamma between cell vectors."""
    alpha = degrees(acos(dot(cell[1], cell[2]) / norm(cell[1]) / norm(cell[2])))
    beta = degrees(acos(dot(cell[0], cell[2]) / norm(cell[0]) / norm(cell[2])))
    gamma = degrees(acos(dot(cell[0], cell[1]) / norm(cell[0]) / norm(cell[1])))
    return alpha, beta, gamma


def xyz2cif(lines):
    """
    Convert xyz file produced by DDEC program to a cif file.

    :param lines: List of lines of XYZ file, e.g. result of open(filename).readlines()
    """
    # Extract important data from the xyz file
    cell, atoms, xyz, charges = xyzparser(lines)

    # Transform cell into a,b,c and alpha, beta, gamma
    cell_a, cell_b, cell_c = cell2abc(cell)
    alpha, beta, gamma = cell2angles(cell)

    # Transform cartensian coordinates into the fractional ones
    fract_xyz = dot(xyz, inv(cell))

    # Create and return a CifFile object
    with tempfile.NamedTemporaryFile(mode='w', suffix='.cif') as file:
        file.write(CIF_HEADER.format(a=cell_a, b=cell_b, c=cell_c, alpha=alpha, beta=beta, gamma=gamma))
        for i, atm in enumerate(atoms):
            file.write(
                '{atm:10} {atm:5} {x:>9.5f} {y:>9.5f} {z:>9.5f} {c:>9.5f}\n'.format(
                    atm=atm, x=fract_xyz[i][0], y=fract_xyz[i][1], z=fract_xyz[i][2], c=charges[i]
                )
            )
        file.flush()
        return CifData(file=file.name, scan_type='flex', parse_policy='lazy')



# -*- coding: utf-8 -*-
"""utils for the workchains"""

from __future__ import absolute_import
import collections
from aiida.engine import workfunction as wf
from aiida.orm import Dict


def dict_merge(dct, merge_dct):
    """ Taken from https://gist.github.com/angstwad/bf22d1822c38a92ec0a9
    Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.
    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None
    """

    for k, _ in merge_dct.items():
        if (k in dct and isinstance(dct[k], dict) and isinstance(merge_dct[k], collections.Mapping)):
            dict_merge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]


@wf
def merge_Dict(p1, p2):  # pylint: disable=invalid-name
    """
    Workfunction to merge two dictionaries p1 and p2 recursively.
    """
    p1_dict = p1.get_dict()
    p2_dict = p2.get_dict()
    dict_merge(p1_dict, p2_dict)
    return Dict(dict=p1_dict).store()


@wf
def extract_core_electrons(cp2k_remote_folder):
    """Read from the cp2k.out the number of core electrons (included in the
    pseudopotential) and print them as a Dict
    """
    with cp2k_remote_folder.creator.outputs.retrieved.open('aiida.out') as handle:
        content = handle.readlines()
    for n_line, line in enumerate(content):
        if '- Atoms:' in line:
            n_atoms = int(line.split()[2])
        if 'Atom  Kind  Element       X           Y           Z' in line:
            break
    res = {
        e.split()[3]: round(float(e.split()[7])) for e in content[n_line + 2:n_line + n_atoms + 2]  # pylint: disable=undefined-loop-variable
    }
    res = [k + ' ' + str(int(k) - int(v)) for k, v in res.items()]
    return Dict(dict={'number of core electrons': res}).store()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Test/example for the DdecCp2kChargesWorkChain"""

from __future__ import absolute_import
from __future__ import print_function
import click
import ase.build

from aiida.plugins import DataFactory, WorkflowFactory
from aiida import cmdline
from aiida import engine
from aiida.orm import Dict, StructureData, SinglefileData
from . import DATA_DIR

Cp2kDdecWorkChain = WorkflowFactory('ddec.cp2k_ddec')  # pylint: disable=invalid-name
StructureData = DataFactory('structure')  # pylint: disable=invalid-name


def run_cp2k_ddec_h2o(cp2k_code, ddec_code):  # pylint: disable=redefined-outer-name
    """Run example for the DdecCp2kChargesWorkChain
    """
    print('Testing CP2K ENERGY calculation + DDEC on H2O...')

    builder = Cp2kDdecWorkChain.get_builder()
    builder.metadata.label = 'test-h2o'

    builder.cp2k_base.cp2k.code = cp2k_code
    atoms = ase.build.molecule('H2O')
    atoms.center(vacuum=2.0)
    builder.cp2k_base.cp2k.structure = StructureData(ase=atoms)
    builder.cp2k_base.cp2k.metadata.options.resources = {
        'num_machines': 1,
        'num_mpiprocs_per_machine': 1,
    }
    builder.cp2k_base.cp2k.metadata.options.withmpi = False
    builder.cp2k_base.cp2k.metadata.options.max_wallclock_seconds = 10 * 60
    builder.cp2k_base.cp2k.parameters = Dict(
        dict={
            'FORCE_EVAL': {
                'METHOD': 'Quickstep',
                'DFT': {
                    'BASIS_SET_FILE_NAME': 'BASIS_MOLOPT',
                    'POTENTIAL_FILE_NAME': 'GTH_POTENTIALS',
                    'QS': {
                        'EPS_DEFAULT': 1.0e-12,
                        'WF_INTERPOLATION': 'ps',
                        'EXTRAPOLATION_ORDER': 3,
                    },
                    'MGRID': {
                        'NGRIDS': 4,
                        'CUTOFF': 280,
                        'REL_CUTOFF': 30,
                    },
                    'XC': {
                        'XC_FUNCTIONAL': {
                            '_': 'LDA',
                        },
                    },
                    'POISSON': {
                        'PERIODIC': 'none',
                        'PSOLVER': 'MT',
                    },
                },
                'SUBSYS': {
                    'KIND': [
                        {
                            '_': 'O',
                            'BASIS_SET': 'DZVP-MOLOPT-SR-GTH',
                            'POTENTIAL': 'GTH-LDA-q6'
                        },
                        {
                            '_': 'H',
                            'BASIS_SET': 'DZVP-MOLOPT-SR-GTH',
                            'POTENTIAL': 'GTH-LDA-q1'
                        },
                    ],
                },
            }
        }
    )
    # The following is not needed, if the files are available in the data directory of your CP2K executable
    cp2k_dir = DATA_DIR / 'cp2k'
    builder.cp2k_base.cp2k.file = {
        'basis': SinglefileData(file=str(cp2k_dir / 'BASIS_MOLOPT')),
        'pseudo': SinglefileData(file=str(cp2k_dir / 'GTH_POTENTIALS')),
    }

    builder.ddec.code = ddec_code
    builder.ddec.parameters = Dict(
        dict={
            'net charge': 0.0,
            'charge type': 'DDEC6',
            'periodicity along A, B, and C vectors': [True, True, True],
            'compute BOs': False,
            'input filename': 'valence_density',
        }
    )
    builder.ddec.metadata.options.max_wallclock_seconds = 10 * 60

    results = engine.run(builder)

    assert 'structure_ddec' in results, results


@click.command()
@cmdline.utils.decorators.with_dbenv()
@click.option('--ddec-code', type=cmdline.params.types.CodeParamType())
@click.option('--cp2k-code', type=cmdline.params.types.CodeParamType())
def cli(ddec_code, cp2k_code):
    """Run example for the DdecCp2kChargesWorkChain
    """
    run_cp2k_ddec_h2o(ddec_code=ddec_code, cp2k_code=cp2k_code)


if __name__ == '__main__':
    cli()  # pylint: disable=no-value-for-parameter



# -*- coding: utf-8 -*-
"""AiiDA-DDEC input plugin"""
from __future__ import absolute_import
import os
from collections import OrderedDict
import six
from voluptuous import Schema, Optional

from aiida.engine import CalcJob
from aiida.orm import Dict, RemoteData, CifData
from aiida.common import CalcInfo, CodeInfo


def input_render(input_dict):
    """
    Converting a dictionary into an input file
    """
    output = ''
    for key, value in input_dict.items():
        if isinstance(value, OrderedDict):
            output += '<' + key + '>' + '\n'
            for k, v in value.items():  # pylint: disable=invalid-name
                output += str(k) + ' ' + str(v) + '\n'
            output += '</' + key + '>' + '\n'
            output += '\n'
        elif isinstance(value, list):
            output += '<' + key + '>' + '\n'
            for e in value:  # pylint: disable=invalid-name
                if e == True:  # pylint: disable=singleton-comparison
                    output += '.true.'
                elif e == False:  # pylint: disable=singleton-comparison
                    output += '.false.'
                else:
                    output += str(e)
                output += '\n'
            output += '</' + key + '>' + '\n'
            output += '\n'
        elif isinstance(value, bool):
            output += '<' + key + '>' + '\n'
            output += '.true.' if value else '.false.'
            output += '\n'
            output += '</' + key + '>' + '\n'
            output += '\n'
        else:
            output += '<' + key + '>' + '\n'
            output += str(value) + '\n'
            output += '</' + key + '>' + '\n'
            output += '\n'
    return output


DENSITY_DIR_KEY = 'atomic densities directory complete path'
DENSITY_DIR_EXTRA = 'DDEC_ATOMIC_DENSITIES_DIRECTORY'
DENSITY_DIR_SYMLINK = 'atomic_densities/'
PARAMETERS_SCHEMA = Schema(
    {
        'net charge': float,
        'charge type': str,
        'periodicity along A, B, and C vectors': [bool, bool, bool],
        'compute BOs': bool,
        Optional(DENSITY_DIR_KEY): str,
        'input filename': str,
        Optional('number of core electrons'): list,
    },
    required=True,
)


def validate_parameters(parameters, port):  # pylint: disable=unused-argument
    """Validate input parameters according to voluptuous schema."""
    PARAMETERS_SCHEMA(parameters.get_dict())


class DdecCalculation(CalcJob):
    """
    AiiDA plugin for the ddec code that performs density derived
    electrostatic and chemical atomic population analysis.
    """

    _DEFAULT_INPUT_FILE = 'job_control.txt'
    _DEFAULT_OUTPUT_FILE = 'valence_cube_DDEC_analysis.output'
    _DEFAULT_ADDITIONAL_RETRIEVE_LIST = '*.xyz'  # pylint: disable=invalid-name

    @classmethod
    def define(cls, spec):
        """
        Init internal parameters at class load time
        """
        # reuse base class function
        super(DdecCalculation, cls).define(spec)
        spec.input(
            'parameters',
            valid_type=Dict,
            validator=validate_parameters,
            help='Input parameters such as net charge, protocol, atomic densities path, ...',
        )
        spec.input(
            'charge_density_folder',
            valid_type=RemoteData,
            required=False,
            help='Use a remote folder (for restarts and similar)',
        )
        spec.inputs['metadata']['options']['parser_name'].default = 'ddec'
        spec.inputs['metadata']['options']['resources'].default = {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 1,
        }
        spec.inputs['metadata']['options']['withmpi'].default = False

        #  exit codes
        spec.exit_code(
            100,
            'ERROR_NO_RETRIEVED_FOLDER',
            message='The retrieved folder data node could not be accessed.',
        )
        spec.exit_code(
            101,
            'ERROR_NO_OUTPUT_FILE',
            message='The retrieved folder does not contain an output file.',
        )
        spec.output('structure_ddec', valid_type=CifData, required=True, help='structure with DDEC charges')

    def prepare_for_submission(self, folder):
        """Create the input files from the input nodes passed
         to this instance of the `CalcJob`.

        :param folder: an `aiida.common.folders.Folder` to temporarily write files on disk
        :return: `aiida.common.datastructures.CalcInfo` instance
        """

        # Determine atomic densities directory
        pm_dict = self.inputs.parameters.get_dict()

        if DENSITY_DIR_KEY not in pm_dict:
            pm_dict[DENSITY_DIR_KEY] = self.inputs.code.extras.get(DENSITY_DIR_EXTRA)

        if not pm_dict[DENSITY_DIR_KEY]:
            raise ValueError(
                f"Please provide '{DENSITY_DIR_KEY}' in the input parameters or set the "
                f"{DENSITY_DIR_EXTRA} extra on the ddec code."
            )

        # The directory must end with a slash or chargemol crashes
        if not pm_dict[DENSITY_DIR_KEY].endswith('/'):
            pm_dict[DENSITY_DIR_KEY] += '/'

        if not os.path.isabs(pm_dict[DENSITY_DIR_KEY]):
            raise ValueError(f"Path to atomic densities directory '{pm_dict[DENSITY_DIR_KEY]}' is not absolute.")

        # Create symlink to atomic densities directory
        density_dir_symlink = (self.inputs.code.computer.uuid, pm_dict[DENSITY_DIR_KEY], DENSITY_DIR_SYMLINK)
        pm_dict[DENSITY_DIR_KEY] = DENSITY_DIR_SYMLINK

        # Write input to file
        input_filename = folder.get_abs_path(self._DEFAULT_INPUT_FILE)
        with open(input_filename, 'w') as infile:
            infile.write(input_render(pm_dict))

        # Prepare CalcInfo to be returned to aiida
        calcinfo = CalcInfo()
        calcinfo.uuid = self.uuid
        calcinfo.local_copy_list = []
        calcinfo.remote_copy_list = []
        calcinfo.remote_symlink_list = [density_dir_symlink]
        calcinfo.retrieve_list = [
            self._DEFAULT_OUTPUT_FILE,
            [self._DEFAULT_ADDITIONAL_RETRIEVE_LIST, '.', 0],
        ]

        # Charge-density remote folder (now working only for CP2K)
        if 'charge_density_folder' in self.inputs:
            charge_density_folder = self.inputs.charge_density_folder
            comp_uuid = charge_density_folder.computer.uuid
            remote_path = os.path.join(
                charge_density_folder.get_remote_path(),
                'aiida-ELECTRON_DENSITY-1_0.cube',
            )
            copy_info = (comp_uuid, remote_path, 'valence_density.cube')
            if self.inputs.code.computer.uuid == comp_uuid:  # if running on the same computer - make a symlink
                calcinfo.remote_symlink_list.append(copy_info)
            else:  # if not - copy the folder
                self.report(
                    f"Warning: Transferring cube file {charge_density_folder.get_remote_path()} from " +
                    f"computer {charge_density_folder.computer.label} to computer {self.inputs.code.computer.label}. " +
                    "This may put strain on your network."
                )
                calcinfo.remote_copy_list.append(copy_info)

        codeinfo = CodeInfo()
        codeinfo.cmdline_params = []
        codeinfo.code_uuid = self.inputs.code.uuid
        calcinfo.codes_info = [codeinfo]

        return calcinfo



# -*- coding: utf-8 -*-
"""Examples of how to run work chains."""
from __future__ import absolute_import
from pathlib import Path

THIS_DIR = Path(__file__).resolve().parent
DATA_DIR = THIS_DIR / 'data'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Setting up DDEC plugin for AiiDA"""
from __future__ import absolute_import
import json
from setuptools import setup, find_packages


def run_setup():
    """Set up aiida-ddec package"""
    with open('setup.json', 'r') as info:
        kwargs = json.load(info)
    setup(
        include_package_data=True,
        packages=find_packages(),
        long_description=open('README.md').read(),
        long_description_content_type='text/markdown',
        **kwargs
    )


if __name__ == '__main__':
    run_setup()
